#include "job.h"

// パラメーターの情報
param_info_t param_info __attribute__((aligned(128)));

// パラメーター
mandel_params_t param __attribute__((aligned(128)));

// SUFのSPEに返す結果
worker_result_t worker_result __attribute__((aligned(128)));


int main(vector unsigned int arg0, vector unsigned int arg1, vector unsigned int arg2)
{
  // このSPEの通し番号
  unsigned int my_num = spu_extract(arg0, 0);
  // 全SPE数（SPSAの分）
  unsigned int spes_num = spu_extract(arg0, 1);
  // データバッファのea
  unsigned int data_ea = spu_extract(arg0, 2);
  // SUFのSPEに返す結果のea
  unsigned int worker_result_ea = data_ea + sizeof(param_info_t) + sizeof(mandel_params_t) + my_num * sizeof(worker_result_t);

  // 完了したジョブのID
  unsigned int done_job = 0;

  unsigned int run = 1;
  while (run)
  {
    // パラメーター情報を読み込む
    mcl_load_and_wait(&param_info, data_ea, sizeof(param_info_t));
    if (param_info.command == COMMAND_WORK)
    {
      if (counter_greater_than(param_info.job_id, done_job))
      {
        // まだ完了していないジョブを実行

        // パラメーターを読み込む
        unsigned int param_ea = data_ea + sizeof(param_info_t);
        mcl_load_and_wait(&param, param_ea, sizeof(mandel_params_t));
        param.para_all = spes_num;
        param.para_my = my_num;

        if (SUF_SPE_WORK)
        {
          param.para_all += NUM_OF_SPE_CLUB;
        }

        job(&param);

        // SUFのSPEに結果を返す(完了したジョブIDを知らせる)
        done_job = param_info.job_id;
        worker_result.job_id = done_job;
        mfc_put(&worker_result, worker_result_ea, sizeof(worker_result_t), 0, 0, 0);
      }
    }
    else if(param_info.command == COMMAND_QUIT)
    {
      // 終了コマンドが来たらループを抜けて終了
      run = 0;
      break;
    }

    // 少し待ってから繰り返す
    mcl_usleep(WAIT_USLEEP);
  }

  return 0;
}
