#include "job.h"

//#define DEBUG

// 独自ストリーム定義パラメーター
#define MY_STREAM_TYPE_ID (spurs::STREAM_TYPE_USER_DEFINE_BEGIN + 0)
#define MY_FIFO_SIZE 128
#define MY_FIFO_TYPE spurs::FIFO_TYPE_SINGLE

const char * name  __attribute__((aligned(128))) = "MyFilter";

// パラメーターの情報
param_info_t param_info __attribute__((aligned(128)));

// パラメーター
mandel_params_t param __attribute__((aligned(128)));

// Worker SPEから返される結果
worker_result_t worker_result[NUM_OF_SPE_SPSA] __attribute__((aligned(128)));

// バッファ
mcl_cf_param_data_t buf_in[32] __attribute__((aligned(128)));
uint32_t buf_h[32] __attribute__((aligned(128)));

// エンディアン変換
static inline unsigned int letobe(unsigned int ui)
{
  vector unsigned char vsp_endian = {3,2,1,0,7,6,5,4,11,10,9,8,15,14,13,12};
  return spu_extract(spu_shuffle(spu_splats(ui), spu_splats(0u), vsp_endian), 0);
}

void info()
{
  suf::def_name(name);
  suf::def_version(1, 0, 0, 0);
  suf::def_parallel_mode(spurs::FILTER_PARALLEL_MODE_SINGLE | spurs::FILTER_PARALLEL_MODE_TILE_DIVISION);

  // 独自ストリーム定義
  suf::def_stream(MY_STREAM_TYPE_ID, MY_FIFO_SIZE, MY_FIFO_TYPE);

  suf::def_input(MY_STREAM_TYPE_ID, 1, 1);
  suf::def_output(spurs::STREAM_TYPE_I420_HD, 1, 1);
}

void init(uint32_t parallel_fraction)
{
}

void fmain(suf::packet_t * input[], suf::packet_t * output[], uint32_t parallel_fraction)
{
  unsigned int param_info_ea, param_ea, worker_result_ea;

  suf::packet_t * in0 = input[0];
  suf::packet_t * out0 = output[0];

  // param_infoをクリア
  param_info.command = 0;
  param_info.job_id = 0;
  param_info.param_ea = 0;

  // ホストからのパラメーターを取得
  suf::dma::get(&buf_in, in0[0].ea, 128);

  // データバッファのメモリ配置について
  // データのバッファ開始ea: (main.cppのbuffer_eaより取得:buf_in[1].ui)
  // |-パラメーターリスト(param_info_t)
  // |-マンデルブロ集合用パラメーター(mandel_params_t)
  // |-Worker SPEから返される結果(worker_result_t) x NUM_OF_SPE_SPSA
  param_info_ea = buf_in[1].ui;
  param_ea = param_info_ea + sizeof(param_info_t);
  worker_result_ea = param_ea + sizeof(mandel_params_t);

  // パケットからパラメーターを読み込んで他のSPEに渡す
  if (in0[0].ea == 0)
  {
    // このパケットは空
    param_info.command = COMMAND_NONE;
    param_info.job_id = 0;
  }
  else
  {
    param_info.command = COMMAND_WORK;
    param_info.job_id = buf_in[0].ui + 1;
    param_info.param_ea = param_ea;

    param.xres = buf_in[2].ui;
    param.yres = buf_in[3].ui;
    param.cx = buf_in[4].f;
    param.cy = buf_in[5].f;
    param.zoom = buf_in[6].f;
    param.frame = buf_in[7].ui;
    param.max_count = buf_in[8].ui;
    param.color_u = buf_in[9].f;
    param.color_v = buf_in[10].f;
    param.out_ea = out0[0].ea;

    // パラメーターをローカルメモリに書き込む
    mfc_put(&param, param_info.param_ea, sizeof(mandel_params_t), 0, 0, 0);
  }

  // パラメーター情報をローカルメモリに書き込む
  // ※フェンスを付けてパラメーターより後に書き込まれるようにする
  // 他のSPEはこれを読み込んで実行開始
  mfc_putf(&param_info, param_info_ea, sizeof(param_info_t), 0, 0, 0);
  mfc_write_tag_mask(1 << 0);
  mfc_write_tag_update_all();
  mfc_read_tag_status();

  if (SUF_SPE_WORK)
  {
    // 自分もジョブを処理する場合
    param.para_all = NUM_OF_SPE_CLUB + NUM_OF_SPE_SPSA;
    // このSPEの番号は最後
    param.para_my = NUM_OF_SPE_CLUB + NUM_OF_SPE_SPSA - 1;
    job(&param);
  }

  // YUVヘッダ情報を書き込み
  spurs::yuv_header_t * yuv_header;
  yuv_header = (spurs::yuv_header_t *)(void *)&buf_h;
  yuv_header->width = spurs::YUV_HEADER_WIDTH_HD_I420;
  yuv_header->height = spurs::YUV_HEADER_HEIGHT_HD_I420;
  yuv_header->horizontal_size = param.xres;
  yuv_header->vertical_size = param.yres;
  yuv_header->picture_structure = spurs::YUV_HEADER_PICTURE_STRUCTURE_PROGRESSIVE;
  mfc_put(&buf_h, param.out_ea, 128, 0, 0, 0);
  mfc_write_tag_mask(1 << 0);
  mfc_write_tag_update_all();
  mfc_read_tag_status();

  // 他のSPEがこのIDのジョブを終わらせるまで待つ
  while (1)
  {
    suf::dma::get(&worker_result, worker_result_ea, sizeof(worker_result_t) * NUM_OF_SPE_SPSA);
    unsigned int not_yet = 0;
    for (int i = 0; i < NUM_OF_SPE_SPSA; i++)
    {
      if (counter_greater_than(param_info.job_id, worker_result[i].job_id))
      {
        not_yet++;
      }
    }

    if (!not_yet)
    {
      break;
    }

    // 少し待ってから繰り返す
    mcl_usleep(WAIT_USLEEP);
  }

  if (in0[0].flags & suf::PACKET_FLAGS_EOS)
  {
    // パケットがEOSだったらEOSをプッシュして終了
    suf::push_eos(0);

    // 終了コマンドを書き込む
    param_info.command = COMMAND_QUIT;
    param_info.job_id = 0;
    mfc_putf(&param_info, param_info_ea, sizeof(param_info_t), 0, 0, 0);
    mfc_write_tag_mask(1 << 0);
    mfc_write_tag_update_all();
    mfc_read_tag_status();
  }

  // パケットをリリース
  suf::release_input(0, 1);
  suf::release_output(0, 1);
}

